//program twobytwo(input,output);
program twobytwo_exact_fishertest;
{
 Copyright (C) Jurg Ott 1988-2008
 Calculates Fisher's exact test and measures of linkage disequilibrium.
 Computes one-sided and two-sided p-values. #locs:387
 Free Pascal. maXbox  #sign:Max: MAXBOX10: 13/09/2019 12:11:30 
 
 { Layout of 2x2 table:

   kk[1]  kk[2] |  m[1]
   kk[3]  kk[4] |  m[2]
   -------------+------
    m[4]   m[3] |   n
 
 History:
  15 Mar 2002  Replace "exact" tests for zero cell frequencies by approx. tests
  02 Jul 2003  Compute only 2-sided p-values
  08 Aug 2003  Output exact p-values even for p < 0.000001
  22 Feb 2006  Display table layout at start-up
  10 Apr 2006  Allow for missing row or column so that program does not
               terminate with error
  04 Sep 2006  Print all tables in critical region, ie. with probabilities
               <= obs table. Add mid-p values.
  20 Oct 2006  Make long output optional
  18 Apr 2007  Minor modifications
  16 Nov 2007  If needed add 1/2 to all cells, not just empty cells
  10 Dec 2007  Minor code changes
  15 Feb 2008  Don't add 1/2 to cells for computing D and D'
  25 Mar 2008  Minor output modifications
  05 Dez 2016  adapt to maXbox4, real to double, direct input
  08 Dez 2016  compare to 0.05 and calc each step to sum 1
}


type
 areal=double; {8-byte reals, double precision}
 ainteger=longint;

const
 aversion = '08 Dezember 2016';
 nmax=100000;
 ndec=6;

type
 vec = array[0..nmax] of real;
 
 //label 42,98,99;
var
 nn,ii:integer;
 kk:array[1..4] of integer;
 rk:array[1..4] of real;
 rn:real;
 LNfac : vec;
 one,half,zero:real;
 longoutput,integeronly:boolean;
 cc:char;
 confv, psumv: double;
 
 function fitoa(aval: double): string;
 begin
   result:= floattoStr(aval)
 end;  


 procedure calcD;
 var
  rr,sd,p1,p2,q1,q2,D,Dmin,Dmax,Dprime,corr:real;
  ii:integer;
  warn:boolean;
 begin
  rn:=zero;
  for ii:=1 to 4 do rn:=rn+rk[ii];
  p1:=(rk[1]+rk[2])/rn;
  p2:=one-p1;
  q1:=(rk[1]+rk[3])/rn;
  q2:=one-q1;
  D:=rk[1]/rn - p1*q1;
  Dmin := -p1*q1;
  rr := -p2*q2;
  if rr>Dmin then Dmin:=rr;
  Dmax:=p1*q2;
  rr:=p2*q1;
  if rr<Dmax then Dmax:=rr;
  writeln('  Linkage disequilibrium:');
 // writeln('  Dmin =',Dmin:10:6,'   Dmax =',Dmax:10:6,'   D =',D:10:6);
  writeln('  Dmin = '+floattostr(Dmin)+'  '+'Dmax = '+floattostr(Dmax)+
                            '   D = '+floattostr(D));
 
  if D>zero then Dprime:=D/Dmax else Dprime:=-D/Dmin;
  //writeln('    D'' ='+,Dprime:10:6);
  writeln('    D''   = '+floattostr(Dprime));

  writeln('  Correlation coefficient:');
  corr:=p1*p2*q1*q2;
  if (corr=0.0)
  then writeln('    undefined (variance = 0)')
  else begin
    corr:= D/sqrt(corr);
    writeln('    r     = '+floattostr(corr));
    corr:=sqr(corr);
 //   writeln('    r^2 =',corr:7:4);
    writeln('    r^2 = '+floattostr(corr));
  end;
  writeln('');
  write('  Odds ratio (approx. relative risk):');
  {Add 1/2 to cells if some cell freq = 0}
  warn:=false;
  for ii:=1 to 4 do if rk[ii]=zero then warn:=true;
  if warn then begin
   for ii:=1 to 4 do rk[ii]:=rk[ii]+half;
   write('  (1/2 added to cells)');
  end;
  writeln('');
  rn:=zero;
  for ii:=1 to 4 do rn:=rn+rk[ii];
  p1:=rk[1]*rk[4]/(rk[2]*rk[3]);
//  writeln('            Estimate = ',p1:8:3);
  writeln('               Estimate = '+floattostr(p1));

  if integeronly then begin
    sd:=zero;
    for ii:=1 to 4 do sd:=sd+one/rk[ii];
    sd:=1.96 * sqrt(sd);
    p1:=ln(p1);
 //   writeln('  95% confid. limits = ',exp(p1-sd):8:3,',',exp(p1+sd):8:3);
    confv:= (exp(p1-sd));
    writeln('  95% confid. limits = '+floattostr(exp(p1-sd))+', '+floattostr(exp(p1+sd)));
  end;
  writeln('');
 end; {calcD}

  procedure exchange(VAR n1,n2:integer);
  var hold:integer;
  begin
   hold:=n1;
   n1:=n2;
   n2:=hold;
  end; {exchange}

 
 procedure calcP;

 { Layout of 2x2 table:

   kk[1]  kk[2] |  m[1]
   kk[3]  kk[4] |  m[2]
   -------------+------
    m[4]   m[3] |   n    }

 var
  ii,jj,mmin,imin,i1,i2:integer;
  m:array[1..4] of integer;
  prob,pcrit,pcrit2,ex,coef,psum, psumtest:real;

  begin {calcP}
  psumtest:= 0.0;
  m[1]:=kk[1]+kk[2];
  m[2]:=kk[3]+kk[4];
  m[3]:=kk[2]+kk[4];
  m[4]:=kk[1]+kk[3];
  nn:=m[1]+m[2];
  LNfac[0]:=zero;
  for ii:=1 to nn do begin
   ex:=ii; {double precision}
   LNfac[ii]:=LN(ex)+LNfac[ii-1];
  end;
  mmin:=nmax;
  for ii:=1 to 4 do if m[ii]<mmin then begin
    mmin:=m[ii];
    imin:=ii;
  end;
  {Smallest marginal should always be #4}
  if imin=1
  then exchange(kk[2],kk[3])
  else if imin=2
   then begin
    exchange(kk[2],kk[3]);
    exchange(kk[1],kk[2]);
    exchange(kk[3],kk[4]);
   end
   else if imin=3
    then begin
     exchange(kk[1],kk[2]);
     exchange(kk[3],kk[4]);
    end;
  {Cell to be increased should be #1}
  ex:=kk[3]+kk[4];
  ex:=(kk[1]+kk[3])*ex/rn; {expected # for k[3]}
  if kk[3]>ex then begin
   exchange(kk[1],kk[3]);
   exchange(kk[2],kk[4]);
  end;

  {Compute p-values}
  psum:=0.0;
  i1:=kk[1];
  i2:=kk[1]+kk[3];

  coef:=LNfac[kk[1]+kk[2]]+LNfac[kk[3]+kk[4]]
    +LNfac[kk[1]+kk[3]]+LNfac[kk[2]+kk[4]]-LNfac[nn];

  if longoutput then begin
   writeln('Tables (possibly rearranged) whose probabilities contribute towards 1p');
   //writeln('Table':20,'Prob':(ndec+4));
   writeln('Table'+'    '+'Prob'+floattostr(ndec+4));
 
  end;
  for ii:=i1 to i2 do begin
   ex:=coef-LNfac[ii]-LNfac[kk[2]]-LNfac[kk[3]]-LNfac[kk[4]];
   prob:=exp(ex);
     //writeln('step '+fitoa(prob))
  
   if longoutput then begin
    write(itoa(ii));
 //   for jj:=2 to 4 do write(kk[jj]:5); write(prob:ndec+4:ndec);
    for jj:=2 to 4 do write(itoa(kk[jj])); write(fitoa(prob)); //:ndec+4:ndec);
    writeln('step '+fitoa(prob))
  
   end;
   psum:=psum+prob;
   if ii=i1 then begin
    pcrit:=prob;
    if longoutput then write(' (obs)');
   end;
   if longoutput then writeln('');
   kk[2]:=kk[2]-1;
   kk[3]:=kk[3]-1;
   kk[4]:=kk[4]+1;
  end; {for ii}
  pcrit2:=half*pcrit;
//  write('  Fisher''s exact test, 1-sided p-value =',psum:ndec+3:ndec);
  write('  Fisher''s exact test, 1-sided p-value = '+fitoa(psum)); //:ndec+3:ndec);
  psumv:= psum;
 {if longoutput then} write(' = '+fitoa(psum));//+fitoa(ndec+5));
  writeln('');
  write('              corresponding mid-p value = '+fitoa(psum-pcrit2));//:ndec+3:ndec);
 {if longoutput then} write(' = '+fitoa(psum))//:(ndec+5));
  writeln('');

 {Now, psum = one-sided significance level. For a two-sided test, one may double this
  value. See discussion in Armitage et al, 4th edition, page 136. Here 2p is
  computed from the total of all tables with prob as small or smaller than the
  observed table. The above establishes the probability of the observed table (pcrit).
  Below, all tables are evaluated and those probabilities are summed up that are less
  than or equal to pcrit}

  writeln('');
  if longoutput then begin
   writeln('Tables whose probabilities contribute towards 2p (including those above)');
 //   writeln('Table':20,'Prob':(ndec+4));
   writeln('Table'+'   Prob'+fitoa(ndec+4));
  end;
  psum:=0;
  for ii:=i2 downto 0 do begin
   kk[2]:=kk[2]+1;
   kk[3]:=kk[3]+1;
   kk[4]:=kk[4]-1;
   ex:=coef-LNfac[ii]-LNfac[kk[2]]-LNfac[kk[3]]-LNfac[kk[4]];
   prob:=exp(ex);
    writeln('step '+itoa(ii)+': '+fitoa(prob))
      psumtest:= psumtest + prob;
   if prob<=pcrit then begin
     //writeln('stepx '+fitoa(prob))
     if longoutput then begin
     write(itoa(ii));
     //for jj:=2 to 4 do write(kk[jj]:5); writeln(prob:ndec+4:ndec);
     for jj:=2 to 4 do write(itoa(kk[jj])); write(fitoa(prob)); //:ndec+4:ndec);
    end;
    psum:=psum+prob;
    //writeln('stepx '+fitoa(psum))
   end;
  end;
   writeln('Sum p: '+fitoa(psumtest))
   writeln('');
 // write('  Fisher''s exact test, 2-sided p-value =',psum:ndec+3:ndec);
  write('  Fisher''s exact test, 2-sided p-value = '+fitoa(psum)); //:ndec+3:ndec);
 {if longoutput then} write(' = '+fitoa(psum)); //:(ndec+5));
  writeln('');
 end; {calcP}


 PROCEDURE showtable;
  BEGIN
   writeln('                                          Risk factor');
   writeln('             Cases Controls        present absent');
   writeln('---------------------------      ----------------------');
   writeln('Risk factor    a      b      or  Cases      a       b');
   writeln(' not at risk   c      d          Controls   c       d');
   writeln('---------------------------      ----------------------');
  END; {showtable}


 procedure init;
 begin
  writeln('Program  2BY2  mX4 version '+(aversion));
  writeln('Maximum number of observations = '+itoa(nmax));
  showtable;
  writeln('');
  writeln('Analyze 2x2 tables by Fisher''s exact test.  For input,');
  writeln('enter the four observations in the 2x2 table');
  zero:=0.0; cc:='S'; one:=1.0; half:=0.5;
 end; {init}


//{$I logo.p}

begin {2by2 main}
 //logo;
 init;
 //repeat
  writeln('');
  //42:
  writeln('Pre Enter the 2 x 2 table below:');
  //if eof then goto 99;
  rn:=0;
  //for ii:=1 to 4 do begin
  // {$I-}  /read(rk[ii]);
   writeln('============= Data input Row: ======================');
   
     { rk[1]:= 5 //4800
      rk[2]:= 0 //4200  
      rk[3]:= 1 //4700 
      rk[4]:= 4 //4100;  //}
 
      rk[1]:= strtofloat(InputBox('F Classifier1-4','Value1','1')) 
      rk[2]:= strtofloat(InputBox('F Classifier2-4','Value2','9'))  
      rk[3]:= strtofloat(InputBox('F Classifier3-4','Value3','11'))
      rk[4]:= strtofloat(InputBox('F Classifier4-4','Value4','3'))  //}
     
   for ii:=1 to 4 do write('  '+fitoa(rk[ii]));
   writeln('============= Data input Section End ==================');
    
   //{$I+}
   if IOresult<>0 then begin
    readln('');
    writeln('== Error data input ==');
    //goto 42;
   end;
   for ii:=1 to 4 do begin
     //if rk[ii]<zero then goto 99;
     rn:=rn+rk[ii];
   end;
  //readln('');
  if rn>nmax then begin
   writeln(' ERROR:  Total n larger than current max of '+itoa(nmax));
   //goto 98;
  end;
  writeln('Short [S] or long [L] output? (blank for S)');
  //if seekeoln then readln else readln(cc);
  longoutput := ((cc='L') or (cc='l'));
  integeronly:=true;
  for ii:=1 to 4 do begin
   kk[ii]:=round(rk[ii]);
   if abs(kk[ii]-rk[ii])>0.0001 then integeronly:=false;
  end;
  if (not integeronly) then longoutput:=true;

  writeln('Input data:');
  if integeronly then
   for ii:=1 to 4 do write(' '+itoa(kk[ii])) else
   //for ii:=1 to 4 do write(rk[ii]:14:6);
   for ii:=1 to 4 do write(fitoa(rk[ii])); //:14:6);
 
  writeln('');

  calcD;  {calculate Delta values}
  if integeronly then calcP;
  writeln('')
  writeln('Fisher Test Result: (Confidence 5%)')
  if (psumv < confv) and (psumv < 0.05) then writeln('H0 reject - Significant!'+' :  '+ 
                fitoa(S_RoundDecimal(psumv,7))+' < '+'  0.05 < '+
                fitoa(S_RoundDecimal(confv,7))) else
                    writeln('H0 accept - NOT significant!');
  writeln('');                  
  writeln('Reference Test of Wiki - en.wikipedia.org/wiki/Fisher''s_exact_test:')
  
     //writeln(floattoStr(binomial(10,1) * binomial(14,11) / binomial(24,12)))
     //writeln(floattoStr(binomial(10,0) * binomial(14,12) / binomial(24,12)))
    writeln('   '+floattoStr(binomial(10,1) * binomial(14,11)/ binomial(24,12) +
                            (binomial(10,0) * binomial(14,12)/ binomial(24,12))));
   writeln('');                    

  //98:
 //until false;
 //99:
End. {twobytwo_fishertest}

Ref: 

                { Layout of 2x2 table:

                  kk[1]  kk[2] |  m[1]
                  kk[3]  kk[4] |  m[2]
                  -------------+------
                   m[4]   m[3] |   n     }

 http://www.jurgott.org/linkage/util.htm
 
 http://www.brixtonhealth.com/miscstat.html
 
 https://en.wikipedia.org/wiki/Fisher's_exact_test
 
                            Men  Women Row total
        Studying 	          1 	9 	   10
        Not-studying   	    11 	3 	   14
        Column total 	    12 	12 	24
        
 Of these only one is more extreme in the same direction as our data; it looks like this:
        
        Studying 	          0 	10 	10
        Non-studying   	   12 	2 	   14
        Column Total 	   12 	12 	24

 Doc:
 In order to calculate the significance of the observed data, i.e. the total probability of observing data as extreme or more extreme if the null hypothesis is true, we have to calculate the values of p for both these tables, and add them together. This gives a one-tailed test, with p approximately 0.001346076 + 0.000033652 = 0.001379728. 
 
 Despite the fact that Fisher's test gives exact p-values, some authors have argued that it is conservative, i.e. that its actual rejection rate is below the nominal significance level.

 //This gives a one-tailed test, with p approximately 0.001346076 + 0.000033652 = 0.001379728. 
   For this table (with extremely unequal studying proportions) the probability is p = ( 10 0 ) ( 14 12 ) / ( 24 12 )  0.000033652 {\displaystyle {p={\tbinom {10}{0}}{\tbinom {14}{12}}}/{\tbinom {24}{12}}\approx 0.000033652} {p = {\tbinom{10}{0}} {\tbinom{14}{12}} }/{ {\tbinom{24}{12}} } \approx 0.000033652.
   
   
   function DSnedecor(Nu1, Nu2 : Integer; X : Float) : Float;
{ Density of Fisher-Snedecor distribution with Nu1 and Nu2 d.o.f. }

 Fisher-Snedecor distribution
  ------------------------------------------------------------------ }
function DSnedecor(Nu1, Nu2 : Integer; X : Float) : Float; external 'dmath';
{ Density of Fisher-Snedecor distribution with Nu1 and Nu2 d.o.f. }
function FSnedecor(Nu1, Nu2 : Integer; X : Float) : Float; external 'dmath';
{ Cumulative prob. for Fisher-Snedecor distrib. with Nu1 and Nu2 d.o.f. }
function PSnedecor(Nu1, Nu2 : Integer; X : Float) : Float; external 'dmath';
{ Prob(F > X) for Fisher-Snedecor distrib. with Nu1 and Nu2 d.o.f. }
function InvSnedecor(Nu1, Nu2 : Integer; P : Float) : Float; external 'dmath';
{ Inverse of Snedecor's F-distribution function }

https://en.wikipedia.org/wiki/Analysis_of_variance

hotline: admin - user -- lan - wlan test: 10,5,6,20 ->0.00764716789355749
online banking: <30 - >50 -- internet - mobile: 10,10,20,5 ->0.0355347

doc:

the typical application of ANOVA, the null hypothesis is that all groups are simply random samples of the same population. For example, when studying the effect of different treatments on similar samples of patients, the null hypothesis would be that all treatments have the same effect (perhaps none). Rejecting the null hypothesis would imply that different treatments result in altered effects.

Examples

## Agresti (1990, p. 61f; 2002, p. 91) Fisher's Tea Drinker
## A British woman claimed to be able to distinguish whether milk or
##  tea was added to the cup first.  To test, she was given 8 cups of
##  tea, in four of which milk was added first.  The null hypothesis
##  is that there is no association between the true order of pouring
##  and the woman's guess, the alternative that there is a positive
##  association (that the odds ratio is greater than 1).
TeaTasting <-
matrix(c(3, 1, 1, 3),
       nrow = 2,
       dimnames = list(Guess = c("Milk", "Tea"),
                       Truth = c("Milk", "Tea")))
fisher.test(TeaTasting, alternative = "greater")
## => p = 0.2429, association could not be established

## Fisher (1962, 1970), Criminal convictions of like-sex twins
Convictions <-
matrix(c(2, 10, 15, 3),
       nrow = 2,
       dimnames =
       list(c("Dizygotic", "Monozygotic"),
            c("Convicted", "Not convicted")))
Convictions
fisher.test(Convictions, alternative = "less")
fisher.test(Convictions, conf.int = FALSE)
fisher.test(Convictions, conf.level = 0.95)$conf.int
fisher.test(Convictions, conf.level = 0.99)$conf.int

Fisher Test Result: (Confidence 5%)
H0 reject - Significant! :  0.0004652 <  0.0056343

## A r x c table  Agresti (2002, p. 57) Job Satisfaction
Job <- matrix(c(1,2,1,0, 3,3,6,1, 10,10,14,9, 6,7,12,11), 4, 4,
dimnames = list(income = c("< 15k", "15-25k", "25-40k", "> 40k"),
                satisfaction = c("VeryD", "LittleD", "ModerateS", "VeryS")))
fisher.test(Job)
fisher.test(Job, simulate.p.value = TRUE, B = 1e5)


"We're not done yet because there is a second source of ambiguity, the calculation of a two-sided p-value. Most people calculate the two-sided p-value by doubling the one-sided p-value. That works perfectly if there is symmetry in the sampling distribution, but that is not the case here. When there is asymmetry, you get a better answer if you sum all the probabilities that are smaller than the probability for the observed table."

Quelle: http://www.pmean.com/11/Fishers.html


https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3884832/table/T2/

http://mathworld.wolfram.com/FishersExactTest.html
http://mathworld.wolfram.com/FishersExactTest.html#eqn2



